@@ -5,6 +5,7 @@ dataset    = 'dataset1'
 start      = '1970-01-01Z'
 stop       = '1970-01-01T00:00:11Z'
 parameters = 'scalar,vector'
-opts       = {'logging': True, 'usecache': False, 'version': 'nightly'}
+#opts       = {'logging': True, 'usecache': False, 'version': 'nightly'}
+opts       = {'logging': True, 'usecache': False }
 
 autoplot(server, dataset, parameters, start, stop, **opts)
