@@ -17,7 +17,7 @@ from hapiplot.plot.util import setopts
 def hapiplot(*args, **kwargs):
     """Plot response from HAPI server.
 
-    Version: 0.2.0
+    Version: 0.2.1
     
     Demos
     -----
@@ -88,7 +88,7 @@ def hapiplot(*args, **kwargs):
 
     Example
     --------
-        >>> server  = 'http://hapi-server.org/servers/TestData/hapi'
+        >>> server  = 'http://hapi-server.org/servers/TestData2.0/hapi'
         >>> dataset = 'dataset1'
         >>> start   = '1970-01-01T00:00:00'
         >>> stop    = '1970-01-02T00:00:00'
@@ -190,7 +190,15 @@ def hapiplot(*args, **kwargs):
     # Convert from NumPy array of byte literals to NumPy array of
     # datetime objects.
     timename = meta['parameters'][0]['name']
-    Time = hapitime2datetime(data[timename])
+
+    nodata = False
+    if len(data[timename]) == 0:
+        nodata = True
+        Time = hapitime2datetime(np.array([meta['x_time.min'], meta['x_time.max']]), allow_missing_Z=True)
+        data = np.ndarray(shape=(2,), dtype=data.dtype)
+        #data[timename] = Time
+    else:
+        Time = hapitime2datetime(data[timename], allow_missing_Z=True)
 
     if len(meta["parameters"]) == 1:
         a = 0 # Time is only parameter
@@ -203,6 +211,116 @@ def hapiplot(*args, **kwargs):
 
         name = meta["parameters"][i]["name"]
 
+        if nodata:
+            # Use NaN values for plot
+            data[name] = np.full(data.dtype[i].shape, np.nan)
+
+
+        name = meta["parameters"][i]["name"]
+
+        if len(data[name].shape) > 3:
+            # TODO: Implement more than 2 dimensions?
+            warning('Parameter ' + name + ' has size with more than 2 dimensions. Plotting first two only.')
+            continue
+
+        # If parameter has a size with two elements, e.g., [N1, N2]
+        # create N2 plots.
+        if len(data[name].shape) == 3:  # shape = (Time, N1, N2)
+
+            log("Parameter '%s' has 3 components. Creating one plot per component." % name, opts)
+
+            pidx = 1 # Primary index is N2
+            sidx = 0 # Secondary index is N1
+            nplts = data[name].shape[1]
+            if data[name].shape[1] > data[name].shape[2]:
+                pidx = 0
+                sidx = 1
+                nplts = data[name].shape[2]
+
+
+            if opts['returnimage']:
+                warning('Only returning first image for parameter with size[1] > 1.')
+                nplts = 1
+            for j in range(nplts):
+                timename = meta['parameters'][0]['name']
+
+                # Name to indicate what is plotted
+
+                if pidx > sidx:
+                    name_new = name + "[" + str(j) + ",:]"
+                else:
+                    name_new = name + "[:," + str(j) + "]"
+
+                # Reduced data ND Array
+                datar = np.ndarray(shape=(data[name].shape[0]),
+                                   dtype=[
+                                           (timename, data.dtype[timename]),
+                                           (name_new, data[name].dtype.str, 
+                                            data[name].shape[1+pidx])
+                                         ])
+
+                datar[timename] = data[timename]
+                if pidx > sidx:
+                    datar[name_new] = data[name][:, j, :]
+                else:
+                    datar[name_new] = data[name][:, :, j]
+
+                # Copy metadata to create a reduced metadata object
+                metar = meta.copy()  # Shallow copy
+                metar["parameters"] = []
+                # Create parameters array with elements of Time parameter ...
+                metar["parameters"].append(meta["parameters"][0])
+                # .... and this parameter
+                metar["parameters"].append(meta["parameters"][i].copy())
+                # Give new name to indicate it is a subset of full parameter
+                metar["parameters"][1]['name'] = name_new
+                metar["parameters"][1]['name_orig'] = name
+
+                # New size is N1 or N2
+                metar["parameters"][1]['size'] = [meta["parameters"][i]['size'][pidx]]
+
+                if 'units' in metar["parameters"][1]:
+                    if type(meta["parameters"][i]['units']) == str or meta["parameters"][i]['units'] == None:
+                        # Same units applies to all dimensions
+                        metar["parameters"][1]["units"] = meta["parameters"][i]['units']
+                    else:
+                        metar["parameters"][1]["units"] = meta["parameters"][i]['units'][j]
+
+                if 'label' in metar["parameters"][1]:
+                    if type(meta["parameters"][i]['label']) == str:
+                        # Same label applies to all dimensions
+                        metar["parameters"][1]["label"] = meta["parameters"][i]['label']
+                    else:
+                        metar["parameters"][1]["label"] = meta["parameters"][i]['label'][j]
+
+                # Extract bins corresponding to jth column of data[name]
+                if 'bins' in metar["parameters"][1]:
+                    metar["parameters"][1]['bins'] = []
+                    metar["parameters"][1]['bins'].append(meta["parameters"][i]['bins'][pidx])
+
+
+                bin_title = ""
+                if 'bins' in meta["parameters"][i] and len(meta["parameters"][i]['bins']) > 1:
+                    bin_title = meta["parameters"][i]['bins'][sidx]['name']
+                    if 'centers' in meta["parameters"][i]['bins'][sidx]:
+                        bin_title = bin_title  + " = " + str(meta["parameters"][i]['bins'][sidx]['centers'][j])
+                        if 'units' in meta["parameters"][i]['bins'][sidx]:
+                            if meta["parameters"][i]['bins'][sidx]['units'] is not None:
+                                bin_title = bin_title + " [" + meta["parameters"][i]['bins'][sidx]['units'] + "]"
+                            bin_title = "\n" + bin_title
+
+                opts['title'] = meta["x_server"] + "\n" + meta["x_dataset"] + " | " + name_new + bin_title
+
+                # rcParams is modified by setopts to have all rcParams.
+                # reset to original passed rcParams so that imagepath
+                # computes file name based on rcParams passed to hapiplot.
+                if 'rcParams' in kwargs:
+                    opts['rcParams'] = kwargs['rcParams']
+
+                metar = hapiplot(datar, metar, **opts)
+                meta["parameters"][i]['hapiplot'] = metar["parameters"][i]['hapiplot']
+            return meta
+
         # Return cached image (case where we are returning binary image data)
         # imagepath() options. Only need filename under these conditions.
         if opts['saveimage'] or (opts['returnimage'] and opts['useimagecache']):
@@ -225,79 +343,8 @@ def hapiplot(*args, **kwargs):
                 meta["parameters"][i]['hapiplot']['image'] = f.read()
             continue
 
-        name = meta["parameters"][i]["name"]
         log("Plotting parameter '%s'" % name, opts)
 
-        if len(data[name].shape) > 3:
-            # TODO: Implement more than 2 dimensions?
-            warning('Parameter ' + name + ' has size with more than 2 dimensions. Plotting first two only.')
-            continue
-
-        # If parameter has a size with two elements, e.g., [N1, N2]
-        # create N2 plots.
-        if len(data[name].shape) == 3:  # shape = (Time, N1, N2)
-
-            nplts = data[name].shape[1]
-            if opts['returnimage']:
-                warning('Only returning first image for parameter with size[1] > 1.')
-                nplts = 1
-            for j in range(nplts):
-                timename = meta['parameters'][0]['name']
-
-                # Name to indicate what is plotted
-                name_new = name + "[:," + str(j) + "]"
-                # Reduced data ND Array
-                datar = np.ndarray(shape=(data[name].shape[0]),
-                                   dtype=[
-                                           (timename, data.dtype[timename]),
-                                           (name_new, data[name].dtype.str, 
-                                            data.dtype[name].shape[1])
-                                         ])
-
-                datar[timename] = data[timename]
-                datar[name_new] = data[name][:, j]
-                # Copy metadata to create a reduced metadata object
-                metar = meta.copy()  # Shallow copy
-                metar["parameters"] = []
-                # Create parameters array with elements of Time parameter ...
-                metar["parameters"].append(meta["parameters"][0])
-                # .... and this parameter
-                metar["parameters"].append(meta["parameters"][i].copy())
-                # Give new name to indicate it is a subset of full parameter
-                metar["parameters"][1]['name'] = name_new
-                metar["parameters"][1]['name_orig'] = name
-                # New size is N1
-                metar["parameters"][1]['size'] = [meta["parameters"][i]['size'][1]]
-
-                if 'units' in metar["parameters"][1]:
-                    if type(meta["parameters"][i]['units']) == str or meta["parameters"][i]['units'] == None:
-                        # Same units applies to all dimensions
-                        metar["parameters"][1]["units"] = meta["parameters"][i]['units']
-                    else:
-                        metar["parameters"][1]["units"] = meta["parameters"][i]['units'][j]
-
-                if 'label' in metar["parameters"][1]:
-                    if type(meta["parameters"][i]['label']) == str:
-                        # Same label applies to all dimensions
-                        metar["parameters"][1]["label"] = meta["parameters"][i]['label']
-                    else:
-                        metar["parameters"][1]["label"] = meta["parameters"][i]['label'][j]
-
-                # Extract bins corresponding to jth column of data[name]
-                if 'bins' in metar["parameters"][1]:
-                    metar["parameters"][1]['bins'] = []
-                    metar["parameters"][1]['bins'].append(meta["parameters"][i]['bins'][j])
-                
-                # rcParams is modified by setopts to have all rcParams.
-                # reset to original passed rcParams so that imagepath
-                # computes file name based on rcParams passed to hapiplot.
-                if 'rcParams' in kwargs:
-                    opts['rcParams'] = kwargs['rcParams']
-
-                metar = hapiplot(datar, metar, **opts)
-                meta["parameters"][i]['hapiplot'] = metar["parameters"][i]['hapiplot']
-            return meta
-
         if opts['title'] != '':
             title = opts['title']
         else:
@@ -347,7 +394,10 @@ def hapiplot(*args, **kwargs):
             if len(name) + len(units) > 30:
                 nl = "\n"
 
-            zlabel = name + nl + " [" + units + "]"
+            #zlabel = name + nl + " [" + units + "]"
+            zlabel = "";
+            if units is not None:
+                zlabel = " [" + units + "]"
 
             if 'bins' in meta['parameters'][i]:
                 if 'ranges' in meta["parameters"][i]['bins'][0]:
@@ -432,14 +482,16 @@ def hapiplot(*args, **kwargs):
                         'backend': opts['backend']
                     }
 
+
             ptype = meta["parameters"][i]["type"]
             if ptype == "isotime":
-                y = hapitime2datetime(data[name])
+                y = hapitime2datetime(data[name],allow_missing_Z=True)
             elif ptype == 'string':
                 y = data[name].astype('U')
             else:
                 y = np.asarray(data[name])
 
+
             if 'fill' in meta["parameters"][i] and meta["parameters"][i]['fill']:
                 if  ptype == 'isotime' or ptype == 'string':
                     Igood = y != meta["parameters"][i]['fill']
@@ -477,7 +529,7 @@ def hapiplot(*args, **kwargs):
 
             if not 'legendlabels' in opts['tsopts']:
                 legendlabels = []
-                if 'size' in meta['parameters'][i]:   
+                if 'size' in meta['parameters'][i]:
                     for l in range(0,meta['parameters'][i]['size'][0]):
                         bin_label = ''
                         bin_name = ''
@@ -499,49 +551,57 @@ def hapiplot(*args, **kwargs):
                             if 'units' in meta['parameters'][i]['bins'][0]:
                                 bin_units = meta['parameters'][i]['bins'][0]['units']
                                 if type(bin_units) == list:
-                                    if type(bin_units[l]) == str:
+                                    if type(bin_units[l]) == str and bin_units != '':
                                         bin_units = ' [' + bin_units[l] + ']'
                                     elif bin_units[l] == None:
-                                        bin_units = ' []'
+                                        bin_units = ' '
                                     else:
-                                        bin_units = ''
+                                        bin_units = ' '
                                 else:
-                                    if type(bin_units) == str:
+                                    if type(bin_units) == str and bin_units != '':
                                        bin_units = ' [' + bin_units + ']'
                                     else:
-                                       bin_units = ''
+                                       bin_units = ' '
+                                
                             if 'centers' in meta['parameters'][i]['bins'][0]:
                                 if meta['parameters'][i]['bins'][0]['centers'][l] is not None:
                                     bin_label = bin_label + ' center = ' + str(meta['parameters'][i]['bins'][0]['centers'][l]) + bin_units
-                                else:
-                                    bin_label = bin_label + ' center = None'
+                                #else:
+                                #   bin_label = bin_label + ' center = None'
 
                             if 'ranges' in meta['parameters'][i]['bins'][0]:
                                 if type(meta['parameters'][i]['bins'][0]['ranges'][l]) == list:
-                                    bin_label = bin_label + sep + ' range = [' + str(meta['parameters'][i]['bins'][0]['ranges'][l][0]) + ', ' + str(meta['parameters'][i]['bins'][0]['ranges'][l][1]) + ']' + bin_units
-                                else:      
-                                    bin_label = bin_label + sep + ' range = [None]'
+                                    if meta['parameters'][i]['bins'][0]['ranges'][l][0] and meta['parameters'][i]['bins'][0]['ranges'][l][1] is not None:
+                                        bin_label = bin_label + sep + ' range = [' + str(meta['parameters'][i]['bins'][0]['ranges'][l][0]) + ', ' + str(meta['parameters'][i]['bins'][0]['ranges'][l][1]) + ']' + bin_units
+                                #else:      
+                                #    bin_label = bin_label + sep + ' range = [None]'
 
                             if bin_label != '':
-                                bin_label = 'bin:' + bin_label
+                                bin_label = 'bin: ' + bin_label
                                 col_name = bin_name + '#%d' % l
 
                         if col_name == '':
                             col_name = 'col #%d' % l
 
+                        if nodata:
+                            col_name = col_name + " [no data in interval]"
+
                         if 'label' in meta['parameters'][i] and \
                             type(meta['parameters'][i]['label']) == list and \
                             len(meta['parameters'][i]['label']) > l and \
                             meta['parameters'][i]['label'][l].strip() != '':
                                 col_name = meta['parameters'][i]['label'][l]
+                                if nodata:
+                                    col_name = col_name + " [no data in interval]"
 
                         if type(units) == list:
                             if len(units) == 1:
-                                legendlabels.append(col_name + ' [' + units[0] + '] ' + bin_label)                                
-                            elif type(units[l]) == str:
+                                if units[0] != '':
+                                    legendlabels.append(col_name + ' [' + units[0] + '] ' + bin_label)
+                            elif type(units[l]) == str and units[l] != '':
                                 legendlabels.append(col_name + ' [' + units[l] + '] ' + bin_label)
                             elif units[l] == None:
-                                legendlabels.append(col_name + ' [] ' + bin_label)
+                                legendlabels.append(col_name + ' ' + bin_label)
                             else:
                                 legendlabels.append(col_name + ' ' + bin_label)
                         else:
@@ -569,7 +629,10 @@ def hapiplot(*args, **kwargs):
             # Apply tsopts
             for key, value in opts['tsopts'].items():
                 tsopts[key] = value
-            
+
+            if nodata == True:
+                tsopts['nodata'] = True
+
             with rc_context(rc=opts['rcParams']):
                 fig = timeseries(Time, y, **tsopts)
 
@@ -606,7 +669,23 @@ def hapiplot(*args, **kwargs):
 
 def imagepath(meta, i, cachedir, opts, fmt):
 
-    optsmd5 = hashlib.md5(json.dumps(opts, sort_keys=True).encode('utf8')).hexdigest()
+    # The value of axis.prop_cycle is a cycler
+    # https://matplotlib.org/cycler/
+    # and can't be serialized by json.dumps.
+    if 'axes.prop_cycle' in opts:
+        opts['axes.prop_cycle'] = list(opts['axes.prop_cycle'])
+
+    try:
+        optsmd5 = hashlib.md5(json.dumps(opts, sort_keys=True).encode('utf8')).hexdigest()
+    except:
+        # Remove elements that can't be serialized.
+        for key in opts:
+            try:
+                json.dumps(opts[key])
+            except:
+                #print('Removed ' + key)
+                opts[key] = None
+        optsmd5 = hashlib.md5(json.dumps(opts, sort_keys=True).encode('utf8')).hexdigest()
 
     fname = request2path(meta['x_server'],
                          meta['x_dataset'],
